@@ -21,6 +21,7 @@
  */
 
 #include <config.h>
+#include "prefs.h"
 #include "ui.h"
 #include "frames.h"
 #include "util.h"
@@ -69,6 +70,7 @@ struct _EventFunc
 {
   MetaEventFunc func;
   gpointer data;
+  int last_event_serial;
 };
 
 static GdkFilterReturn
@@ -83,7 +85,10 @@ filter_func (GdkXEvent *xevent,
   if ((* ef->func) (xevent, ef->data))
     return GDK_FILTER_REMOVE;
   else
-    return GDK_FILTER_CONTINUE;
+    {
+      ef->last_event_serial = ((XEvent*)xevent)->xany.serial;
+      return GDK_FILTER_CONTINUE;
+    }
 }
 
 static EventFunc *ef = NULL;
@@ -116,6 +121,16 @@ meta_ui_remove_event_func (Display       *xdisplay,
   ef = NULL;
 }
 
+int
+meta_ui_get_last_event_serial (Display *xdisplay)
+{
+  g_assert (ef != NULL);
+
+  /* This is technically broken since it's not per-display */
+  
+  return ef->last_event_serial;
+}
+
 MetaUI*
 meta_ui_new (Display *xdisplay,
              Screen  *screen)
@@ -154,7 +169,8 @@ meta_ui_get_frame_geometry (MetaUI *ui,
 
 Window
 meta_ui_create_frame_window (MetaUI *ui,
-			     Display *xdisplay,
+                             Display *xdisplay,
+                             Visual *xvisual,
 			     gint x,
 			     gint y,
 			     gint width,
@@ -166,12 +182,22 @@ meta_ui_create_frame_window (MetaUI *ui,
   GdkWindowAttr attrs;
   gint attributes_mask;
   GdkWindow *window;
+  GdkVisual *visual;
+  GdkColormap *cmap = gdk_screen_get_default_colormap (screen);
   
   /* Default depth/visual handles clients with weird visuals; they can
    * always be children of the root depth/visual obviously, but
    * e.g. DRI games can't be children of a parent that has the same
    * visual as the client.
    */
+  if (!xvisual)
+    visual = gdk_screen_get_system_visual (screen);
+  else
+    {
+      visual = gdk_x11_screen_lookup_visual (screen,
+                                             XVisualIDFromVisual (xvisual));
+      cmap = gdk_colormap_new (visual, FALSE);
+    }
 
   attrs.title = NULL;
 
@@ -185,8 +211,8 @@ meta_ui_create_frame_window (MetaUI *ui,
   attrs.x = x;
   attrs.y = y;
   attrs.wclass = GDK_INPUT_OUTPUT;
-  attrs.visual = gdk_screen_get_system_visual (screen);
-  attrs.colormap = gdk_screen_get_default_colormap (screen);
+  attrs.visual = visual;
+  attrs.colormap = cmap;
   attrs.window_type = GDK_WINDOW_CHILD;
   attrs.cursor = NULL;
   attrs.wmclass_name = NULL;
@@ -659,6 +685,44 @@ meta_text_property_to_utf8 (Display             *xdisplay,
   return retval;
 }
 
+void
+meta_ui_theme_get_frame_borders (MetaUI *ui,
+                                 MetaFrameType      type,
+                                 MetaFrameFlags     flags,
+                                 int               *top_height,
+                                 int               *bottom_height,
+                                 int               *left_width,
+                                 int               *right_width)
+{
+  int text_height;
+  PangoContext *context;
+  const PangoFontDescription *font_desc;
+  GtkStyle *default_style;
+
+  if (meta_ui_have_a_theme ())
+    {
+      context = gtk_widget_get_pango_context (GTK_WIDGET (ui->frames));
+      font_desc = meta_prefs_get_titlebar_font ();
+
+      if (!font_desc)
+        {
+          default_style = gtk_widget_get_default_style ();
+          font_desc = default_style->font_desc;
+        }
+
+      text_height = meta_pango_font_desc_get_text_height (font_desc, context);
+
+      meta_theme_get_frame_borders (meta_theme_get_current (),
+                                    type, text_height, flags,
+                                    top_height, bottom_height,
+                                    left_width, right_width);
+    }
+  else
+    {
+      *top_height = *bottom_height = *left_width = *right_width = 0;
+    }
+}
+
 void
 meta_ui_set_current_theme (const char *name,
                            gboolean    force_reload)
